from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import fixture

import piculet

import logging
import os
import socket


logging.raiseExceptions = False


cache_dir = os.path.join(os.path.dirname(__file__), '.cache')
os.environ['PICULET_WEB_CACHE'] = cache_dir
# os.environ['HTTPS_PROXY'] = 'http://localhost:8123'

enabled_socket = socket.socket


TEST_PAGES = {
    'bowie': 'https://en.wikipedia.org/wiki/David_Bowie'
}


for page in TEST_PAGES.values():
    piculet.get_document(page)


@fixture(scope='session')
def test_pages():
    """Addresses and cache locations of test pages."""
    return {k: (v, os.path.join(cache_dir, piculet.get_hash(v)))
            for k, v in TEST_PAGES.items()}


@fixture(scope='session', autouse=True)
def disable_internet(request):
    """Disable Internet access."""

    def disabled_socket(*args, **kwargs):
        raise RuntimeError('Internet access disabled')

    old_socket = socket.socket
    socket.socket = disabled_socket
    yield
    socket.socket = old_socket


@fixture(scope='function')
def enable_internet(request):
    """Enable Internet access."""
    old_socket = socket.socket
    socket.socket = enabled_socket
    yield
    socket.socket = old_socket



from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import fixture, raises

import os

from piculet import build_tree, extract, reducers


shining_file = os.path.join(os.path.dirname(__file__), '..', 'examples', 'shining.html')
shining_content = open(shining_file).read()


@fixture
def shining():
    """Root node of the XML tree for the movie document "The Shining"."""
    return build_tree(shining_content)


def test_preprocess_unknown_operation_should_raise_error(shining):
    with raises(ValueError):
        extract(shining, [], pre=[{"op": "foo"}])


def test_preprocess_root_multiple_should_raise_error(shining):
    with raises(ValueError):
        extract(shining, [], pre=[{"op": "root", "path": ".//div"}])


def test_preprocess_root_none_should_return_empty_result(shining):
    items = [{"key": "title", "value": {"path": ".//title/text()", "reduce": reducers.first}}]
    data = extract(shining, items, pre=[{"op": "root", "path": ".//foo"}])
    assert data == {}


def test_preprocess_root_should_exclude_unselected_parts(shining):
    items = [{"key": "foo", "value": {"path": ".//h3/text()", "reduce": reducers.first}}]
    data = extract(shining, items, pre=[{"op": "root", "path": ".//div[@class='director']"}])
    assert data == {'foo': 'Director:'}


def test_preprocess_remove_should_remove_selected_node(shining):
    items = [{"key": "cast",
              "value": {
                  "foreach": ".//table[@class='cast']/tr",
                  "items": [
                      {"key": "name",
                       "value": {"path": "./td[1]/a/text()",
                                 "reduce": reducers.first}}
                  ]
              }}]
    data = extract(shining, items, pre=[{"op": "remove", "path": ".//tr[1]"}])
    assert data == {'cast': [{'name': 'Shelley Duvall'}]}


def test_preprocess_remove_selected_none_should_not_cause_error(shining):
    items = [{"key": "cast",
              "value": {
                  "foreach": ".//table[@class='cast']/tr",
                  "items": [
                      {"key": "name",
                       "value": {"path": "./td[1]/a/text()",
                                 "reduce": reducers.first}}
                  ]
              }}]
    data = extract(shining, items, pre=[{"op": "remove", "path": ".//tr[50]"}])
    assert data == {'cast': [{'name': 'Jack Nicholson'}, {'name': 'Shelley Duvall'}]}



# Copyright (C) 2014-2017 H. Turgut Uyar <uyar@tekir.org>
#
# Piculet is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Piculet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Piculet.  If not, see <http://www.gnu.org/licenses/>.

"""
Piculet is a package for extracting data from XML documents using XPath queries.
It can also scrape web pages by first converting the HTML source into XHTML.
Piculet consists of this single source file with no dependencies other than
the standard library, which makes it very easy to integrate into applications.

For more details, please refer to the documentation: https://piculet.readthedocs.io/
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from argparse import ArgumentParser
from collections import deque
from functools import partial
from hashlib import md5
from operator import itemgetter
from pkgutil import find_loader

import json
import logging
import os
import re
import sys


PY3 = sys.version_info >= (3, 0)
PY33 = sys.version_info >= (3, 3)
PY34 = sys.version_info >= (3, 4)


if not PY3:
    str, bytes = unicode, str


if not PY3:
    from cgi import escape as html_escape
    from HTMLParser import HTMLParser
    from StringIO import StringIO
    from urllib2 import urlopen
else:
    from html import escape as html_escape
    from html.parser import HTMLParser
    from io import StringIO
    from urllib.request import urlopen


if not PY33:
    class SimpleNamespace:
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

        def __eq__(self, other):
            return self.__dict__ == other.__dict__
else:
    from types import SimpleNamespace


if not PY34:
    from contextlib import contextmanager

    @contextmanager
    def redirect_stdout(new_stdout):
        old_stdout, sys.stdout = sys.stdout, new_stdout
        try:
            yield new_stdout
        finally:
            sys.stdout = old_stdout
else:
    from contextlib import redirect_stdout


_logger = logging.getLogger(__name__)


###########################################################
# HTML OPERATIONS
###########################################################


# TODO: this is too fragile
_CHARSET_TAGS = {
    'meta charset': b'<meta charset="',
    'meta http-equiv': b'<meta http-equiv="content-type" content="text/html; charset='
}


def decode_html(content, charset=None, fallback_charset='utf-8'):
    """Decode an HTML document according to a character set.

    If no character set is given, this will try to figure it out
    from the corresponding ``meta`` tags.

    :sig: (bytes, Optional[str], Optional[str]) -> str
    :param content: Content of HTML document to decode.
    :param charset: Character set of the page.
    :param fallback_charset: Character set to use if it can't be figured out.
    :return: Decoded content of the document.
    """
    if charset is None:
        for key, tag in _CHARSET_TAGS.items():
            start = content.find(tag)
            if start >= 0:
                _logger.debug('charset found in [%s] tag', key)
                charset_start = start + len(tag)
                charset_end = content.find(b'"', charset_start)
                charset = content[charset_start:charset_end].decode('ascii')
                break
        else:
            _logger.debug('charset not found, using fallback')
            charset = fallback_charset
    _logger.debug('decoding for charset [%s]', charset)
    return content.decode(charset)


class HTMLNormalizer(HTMLParser):
    """HTML cleaner and XHTML convertor.

    DOCTYPE declarations and comments are removed.

    :sig: (Set[str], Set[str]) -> None
    :param omit_tags: Tags to remove, along with all their content.
    :param omit_attrs: Attributes to remove.
    """

    SELF_CLOSING_TAGS = {'br', 'hr', 'img', 'input', 'link', 'meta'}
    """Tags to handle as self-closing."""

    def __init__(self, omit_tags=(), omit_attrs=()):
        if not PY3:
            HTMLParser.__init__(self)
        elif not PY34:
            super().__init__()
        else:
            super().__init__(convert_charrefs=True)

        self.omit_tags = set(omit_tags)     # sig: Set[str]
        self.omit_attrs = set(omit_attrs)   # sig: Set[str]

        # stacks used during normalization
        self._open_tags = deque()
        self._open_omitted_tags = deque()

    def handle_starttag(self, tag, attrs):
        if tag in self.omit_tags:
            _logger.debug('omitting [%s] tag', tag)
            self._open_omitted_tags.append(tag)
        if not self._open_omitted_tags:
            # stack empty -> not in omit mode
            if (tag == 'li') and (self._open_tags[-1] == 'li'):
                _logger.debug('opened [li] without closing previous [li], adding closing tag')
                self.handle_endtag('li')
            attribs = []
            for attr_name, attr_value in attrs:
                if attr_name in self.omit_attrs:
                    _logger.debug('omitting [%s] attribute of [%s] tag', attr_name, tag)
                    continue
                if attr_value is None:
                    _logger.debug('no value for [%s] attribute of [%s] tag, adding empty value',
                                  attr_name, tag)
                    attr_value = ''
                attribs.append((attr_name, attr_value))
            attrs = ['%(name)s="%(value)s"' % {
                'name': name,
                'value': html_escape(value, quote=True)
            } for name, value in attribs]
            line = '<%(tag)s%(attrs)s%(slash)s>' % {
                'tag': tag,
                'attrs': (' ' + ' '.join(attrs)) if len(attrs) > 0 else '',
                'slash': ' /' if tag in self.SELF_CLOSING_TAGS else ''
            }
            print(line, end='')
            if tag not in self.SELF_CLOSING_TAGS:
                self._open_tags.append(tag)

    def handle_endtag(self, tag):
        if not self._open_omitted_tags:
            # stack empty -> not in omit mode
            if tag not in self.SELF_CLOSING_TAGS:
                last = self._open_tags[-1]
                if (tag == 'ul') and (last == 'li'):
                    _logger.debug('closing [ul] without closing last [li], adding closing tag')
                    self.handle_endtag('li')
                if tag == last:
                    # expected end tag
                    print('</%(tag)s>' % {'tag': tag}, end='')
                    self._open_tags.pop()
                elif tag not in self._open_tags:
                    _logger.debug('end tag [%s] without start tag', tag)
                    # XXX: for <a><b></a></b>, this case gets invoked after the case below
                elif tag == self._open_tags[-2]:
                    _logger.debug('unexpected end tag [%s] instead of [%s], closing both',
                                  tag, last)
                    print('</%(tag)s>' % {'tag': last}, end='')
                    print('</%(tag)s>' % {'tag': tag}, end='')
                    self._open_tags.pop()
                    self._open_tags.pop()
        elif (tag in self.omit_tags) and (tag == self._open_omitted_tags[-1]):
            # end of expected omitted tag
            self._open_omitted_tags.pop()

    def handle_data(self, data):
        if not self._open_omitted_tags:
            # stack empty -> not in omit mode
            print(html_escape(data), end='')

    def handle_entityref(self, name):
        # XXX: doesn't get called if convert_charrefs=True
        ref = self.unescape('&' + name + ';')
        print('&#%d;' % ord(ref))

    def handle_charref(self, name):
        # XXX: doesn't get called if convert_charrefs=True
        print('&#' + name + ';', end='')

    # def feed(self, data):
        # super().feed(data)
        # close all remaining open tags
        # for tag in reversed(self._open_tags):
        #     print('</%(tag)s>' % {'tag': tag}, end='')


def html_to_xhtml(document, omit_tags=(), omit_attrs=()):
    """Clean HTML and convert to XHTML.

    :sig: (str, Optional[Iterable[str]], Optional[Iterable[str]]) -> str
    :param document: HTML document to clean and convert.
    :param omit_tags: Tags to exclude from the output.
    :param omit_attrs: Attributes to exclude from the output.
    :return: Normalized XHTML content.
    """
    _logger.debug('cleaning html and converting to xhtml')
    out = StringIO()
    normalizer = HTMLNormalizer(omit_tags=omit_tags, omit_attrs=omit_attrs)
    with redirect_stdout(out):
        normalizer.feed(document)
    return out.getvalue()


###########################################################
# DATA EXTRACTION OPERATIONS
###########################################################


_USE_LXML = find_loader('lxml') is not None
if _USE_LXML:
    from lxml import etree as ElementTree
    _logger.debug('using lxml')
else:
    from xml.etree import ElementTree
    _logger.debug('lxml not found, falling back to elementtree')


def build_tree(document):
    """Build a tree from an XML document.

    :sig: (str) -> ElementTree.Element
    :param document: XML document to build the tree from.
    :return: Root node of the XML tree.
    """
    return ElementTree.fromstring(document if PY3 else document.encode('utf-8'))


def xpath_etree(element, path):
    """Apply an XPath expression to an element.

    This function is mainly needed to compensate for the lack of ``text()``
    and ``@attr`` axis queries in ElementTree XPath support.

    :sig: (ElementTree.Element, str) -> Union[List[str], List[ElementTree.Element]]
    :param element: Element to apply the expression to.
    :param path: XPath expression to apply.
    :return: Elements or strings resulting from the query.
    """
    if path.endswith('//text()'):
        _logger.debug('handling descendant text path [%s]', path)
        return [t for e in element.findall(path[:-8]) for t in e.itertext() if t]

    if path.endswith('/text()'):
        _logger.debug('handling child text path [%s]', path)
        return [t for e in element.findall(path[:-7])
                for t in ([e.text] + [c.tail if c.tail else '' for c in e]) if t]

    path_tokens = path.split('/')
    epath, last_step = path_tokens[:-1], path_tokens[-1]
    # PY3: *epath, last_step = path.split('/')
    if last_step.startswith('@'):
        _logger.debug('handling attribute path [%s]', path)
        result = [e.attrib.get(last_step[1:]) for e in element.findall('/'.join(epath))]
        return [r for r in result if r is not None]

    _logger.debug('handling element producing path [%s]', path)
    return element.findall(path)


xpath = xpath_etree if not _USE_LXML else ElementTree._Element.xpath


reducers = SimpleNamespace(
    first=itemgetter(0),
    join=partial(str.join, ''),
    clean=lambda xs: re.sub('\s+', ' ', ''.join(xs).replace('\xa0', ' ')).strip(),
    normalize=lambda xs: re.sub('[^a-z0-9_]', '', ''.join(xs).lower().replace(' ', '_'))
)
"""Predefined reducers."""


def woodpecker(path, reduce=None, reducer=None):
    """A value extractor that combines an XPath query with a reducing function.

    This function returns a callable that takes an XML element as parameter and
    applies the XPath query to that element to get a list of strings; therefore
    the query has to end with ``text()`` or ``@attr``. The list will then be passed
    to the reducing function to generate a single string as the result.

    Either the ``reduce`` parameter must be supplied as a callable,
    or the ``reducer`` must be supplied as the name of a predefined reducer function.
    If both are supplied, the ``reduce`` parameter will be used.

    :sig:
        (
            str,
            Optional[Callable[[List[str]], str]],
            Optional[str]
        ) -> Callable[[ElementTree.Element], Optional[str]]
    :param path: XPath query to select the values.
    :param reduce: Function to reduce the selected elements to a single value.
    :param reducer: Name of predefined reducer function.
    :return: A callable that can apply this path and reducer to an element.
    """
    if reduce is None:
        if reducer is None:
            raise ValueError('A reducing function must be specified')
        try:
            reduce = getattr(reducers, reducer)
        except AttributeError:
            raise ValueError('Unknown reducer: ' + reducer)

    def apply(element):
        """Extract a value from an element.

        :param element: Element to extract the value from.
        :return: Extracted value.
        """
        values = xpath(element, path)
        if len(values) == 0:
            _logger.debug('no match for [%s]', path)
            return None

        _logger.debug('extracted value [%s]', values)
        value = reduce(values)
        _logger.debug('applied [%s] reducer, new value [%s]', reducer, value)
        return value

    return apply


def extract(root, items, pre=()):
    """Extract data from an XML tree.

    This will extract the data items according to the supplied rules.
    If given, it will use the ``pre`` parameter to carry out preprocessing
    operations on the tree before starting data extraction.

    :sig:
        (
            ElementTree.Element,
            Iterable[Mapping[str, Any]],
            Optional[Iterable[Mapping[str, Any]]]
        ) -> Mapping[str, Any]
    :param root: Root of the XML tree to extract the data from.
    :param items: Rules that specify how to extract the data items.
    :param pre: Preprocessing operations on the document tree.
    :return: Extracted data.
    """

    def gen(val):
        """Value generator function.

        It takes a value generator description and returns a callable
        that takes an XML element and returns a value.
        """
        if isinstance(val, str):
            # constant function
            return lambda _: val
        if 'path' in val:
            # xpath and reducer
            if ('reduce' not in val) and ('reducer' not in val):
                raise ValueError('Path extractor must have reducer: ' + val['path'])
            return woodpecker(val['path'], reduce=val.get('reduce'),
                              reducer=val.get('reducer'))
        if 'items' in val:
            # recursive function for applying subrules
            return partial(extract, items=val['items'], pre=val.get('pre', ()))
        raise ValueError('Unknown value generator: ' + str(val))

    # ElementTree doesn't support parent queries, so build a map for it
    # TODO: this re-traverses tree on subrules
    get_parent = {e: p for p in root.iter() for e in p}.get if not _USE_LXML else \
        ElementTree._Element.getparent

    # do the preprocessing operations
    for step in pre:
        op = step['op']
        if op == 'root':
            path = step['path']
            _logger.debug('selecting new root using [%s]', path)
            elements = xpath(root, path)
            if len(elements) > 1:
                raise ValueError('Root expression must not select multiple elements: ' + path)
            if len(elements) == 0:
                _logger.debug('no matches for new root, no data to extract')
                return {}
            root = elements[0]
        elif op == 'remove':
            path = step['path']
            elements = xpath(root, path)
            _logger.debug('removing %s elements using [%s]', len(elements), path)
            for element in elements:
                # XXX: could this be hazardous? parent removed in earlier iteration?
                get_parent(element).remove(element)
        elif op == 'set_attr':
            path = step['path']
            attr_name_gen = gen(step['name'])
            attr_value_gen = gen(step['value'])
            for element in xpath(root, path):
                attr_name = attr_name_gen(element)
                attr_value = attr_value_gen(element)
                _logger.debug('setting [%s] attribute to [%s] on [%s] element',
                              attr_name, attr_value, element.tag)
                element.attrib[attr_name] = attr_value
        elif op == 'set_text':
            path = step['path']
            text_gen = gen(step['text'])
            for element in xpath(root, path):
                text_value = text_gen(element)
                _logger.debug('setting text to [%s] on [%s] element', text_value, element.tag)
                element.text = text_value
        else:
            raise ValueError('Unknow preprocessing operation: ' + op)

    # actual extraction part
    data = {}
    for item in items:
        key_gen = gen(item['key'])
        foreach_key = item.get('foreach')
        subroots = [root] if foreach_key is None else xpath(root, foreach_key)
        for subroot in subroots:
            raw_key = key_gen(subroot)
            try:
                key_transform = item['key'].get('transform')
            except AttributeError:
                key_transform = None
            key = raw_key if key_transform is None else key_transform(raw_key)

            value_gen = gen(item['value'])
            foreach_value = item['value'].get('foreach')
            transform = item['value'].get('transform')
            if foreach_value is None:
                raw_value = value_gen(subroot)
                if raw_value:   # None from pecker, or {} from extract
                    # XXX: consider '', 0, and other non-truthy values
                    data[key] = raw_value if transform is None else transform(raw_value)
            else:
                values = [value_gen(r) for r in xpath(subroot, foreach_value)]
                if values:
                    data[key] = values if transform is None else [transform(v) for v in values]
    return data


def get_hash(url):
    """Get the hash value of a URL for cache lookups.

    :sig: (str) -> str
    :param url: URL to compute the hash for.
    :return: Hash value of URL.
    """
    return md5(url.encode('utf-8')).hexdigest()


def get_document(url):
    """Get the document with the given URL.

    This function will check whether the requested document has already been retrieved
    and return the cached content if possible. Note that cached documents are assumed
    to be up-to-date.

    :sig: (str) -> str
    :param url: URL to get the document from.
    :return: Content of the retrieved document.
    """
    _logger.debug('getting page [%s]', url)
    cache_dir = os.environ.get('PICULET_WEB_CACHE')
    if cache_dir is None:
        _logger.debug('cache disabled, downloading page')
        content = urlopen(url).read()
    else:
        _logger.debug('using cache dir [%s]', cache_dir)
        if not PY3:
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
        else:
            os.makedirs(cache_dir, exist_ok=True)
        key = get_hash(url)
        cache_file = os.path.join(cache_dir, key)
        if not os.path.exists(cache_file):
            _logger.debug('page not in cache, downloading and storing')
            content = urlopen(url).read()
            with open(cache_file, 'wb') as f:
                f.write(content)
        else:
            _logger.debug('using page found in cache')
            with open(cache_file, 'rb') as f:
                content = f.read()
    _logger.debug('read %s bytes', len(content))
    return decode_html(content)


def scrape(document, rules, content_format='xml'):
    """Extract data from a document according to given rules.

    :sig: (str, Mapping[str, Any], Optional[str]) -> Mapping[str, Any]
    :param document: Document to scrape.
    :param rules: Rules to use for scraping.
    :param content_format: Format of the content, XML or HTML.
    :return: Extracted data.
    """
    if content_format == 'html':
        _logger.debug('converting html document to xml')
        document = html_to_xhtml(document)
        # _logger.debug('=== CONTENT START ===\n%s\n=== CONTENT END===', document)
    root = build_tree(document)
    data = extract(root, rules['items'], pre=rules.get('pre', ()))
    return data


###########################################################
# COMMAND-LINE INTERFACE
###########################################################


def h2x(source):
    """Convert an HTML file into XHTML and print.

    :sig: (str) -> None
    :param source: Path of HTML file to convert.
    """
    _logger.debug('converting html to xhtml')
    if source == '-':
        _logger.debug('reading from stdin')
        content = sys.stdin.read()
    else:
        path = os.path.abspath(source)
        _logger.debug('reading from file [%s]', path)
        with open(path, 'rb') as f:
            content = decode_html(f.read())
    print(html_to_xhtml(content), end='')


def scrape_document(address, spec, content_format='xml'):
    """Scrape data from a URL and print.

    :sig: (str, str, Optional[str]) -> None
    :param address: File path or URL of document to scrape.
    :param spec: Path of spec file.
    :param content_format: Whether the content is XML or HTML.
    """
    with open(spec) as f:
        rules = json.loads(f.read())
    if address.startswith(('http://', 'https://')):
        _logger.debug('scraping url [%s]', address)
        document = get_document(address)
    else:
        _logger.debug('scraping file [%s]', address)
        with open(address, 'rb') as f:
            document = decode_html(f.read())
    data = scrape(document, rules, content_format=content_format)
    print(json.dumps(data, indent=2, sort_keys=True))


def _get_parser(prog):
    """Get a parser for command line arguments."""

    def _h2x(arguments):
        h2x(arguments.file)

    def _scrape(arguments):
        content_format = 'html' if arguments.html else 'xml'
        scrape_document(arguments.document, arguments.spec, content_format=content_format)

    parser = ArgumentParser(prog=prog)
    parser.add_argument('--debug', action='store_true', help='enable debug messages')

    commands = parser.add_subparsers(metavar='command', dest='command')
    commands.required = True

    subparser = commands.add_parser('h2x', help='convert HTML to XHTML')
    subparser.set_defaults(func=_h2x)
    subparser.add_argument('file', help='file to convert')

    subparser = commands.add_parser('scrape', help='scrape a document')
    subparser.set_defaults(func=_scrape)
    subparser.add_argument('document', help='file path or URL of document to scrape')
    subparser.add_argument('-s', '--spec', required=True, help='spec file')
    subparser.add_argument('--html', action='store_true', help='document is in HTML format')

    return parser


def main(argv=None):
    """Entry point of the command line utility.

    :sig: (Optional[List[str]]) -> None
    :param argv: Command line arguments.
    """
    argv = argv if argv is not None else sys.argv
    parser = _get_parser(prog=argv[0])
    arguments = parser.parse_args(argv[1:])

    # set debug mode
    if arguments.debug:
        logging.basicConfig(level=logging.DEBUG)
        _logger.debug('running in debug mode')

    # run the handler for the selected command
    try:
        arguments.func(arguments)
    except Exception as e:
        print(e, file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import raises

import os

from piculet import build_tree, extract, reducers, woodpecker, xpath


########################################
# xpath tests
########################################


generic_content = '<d><t1>foo</t1><t1 a="v"><t2>bar</t2></t1></d>'
generic_root = build_tree(generic_content)


def test_xpath_non_text_queries_should_return_nodes():
    selected = xpath(generic_root, ".//t1")
    assert [s.tag for s in selected] == ['t1', 't1']


def test_xpath_child_text_queries_should_return_strings():
    selected = xpath(generic_root, ".//t1/text()")
    assert selected == ['foo']


def test_xpath_descendant_text_queries_should_return_strings():
    selected = xpath(generic_root, ".//t1//text()")
    assert selected == ['foo', 'bar']


def test_xpath_attr_queries_should_return_strings():
    selected = xpath(generic_root, ".//t1/@a")
    assert selected == ['v']


########################################
# reducer tests
########################################


def test_reducer_first_should_return_first_item():
    assert reducers.first(['a', 'b', 'c']) == 'a'


def test_reducer_join_should_return_joined_items():
    assert reducers.join(['a', 'b', 'c']) == 'abc'


def test_reducer_clean_should_remove_extra_space():
    assert reducers.clean(['  a ', '   b', ' c ']) == 'a b c'


def test_reducer_clean_should_treat_nbsp_as_space():
    assert reducers.clean(['  a ', ' \xa0  b', ' c ']) == 'a b c'


def test_reducer_normalize_should_convert_to_lowercase():
    assert reducers.normalize(['A', 'B', 'C']) == 'abc'


def test_reducer_normalize_should_remove_nonalphanumeric_characters():
    assert reducers.normalize(['a+', '?b7', '{c}']) == 'ab7c'


def test_reducer_normalize_should_keep_underscores():
    assert reducers.normalize(['a_', 'b', 'c']) == 'a_bc'


def test_reducer_normalize_should_replace_spaces_with_underscores():
    assert reducers.normalize(['a', ' b', 'c']) == 'a_bc'



########################################
# woodpecker tests
########################################


people_content = '<p><p1><n>John Smith</n><a>42</a></p1><p2><n>Jane Doe</n></p2></p>'
people_root = build_tree(people_content)


def test_peck_no_reducer_should_raise_error():
    with raises(ValueError):
        woodpecker(path=".//n/text()")


def test_peck_unknown_predefined_reducer_should_raise_error():
    with raises(ValueError):
        woodpecker(path=".//n/text()", reducer="merge")


def test_peck_known_predefined_reducer_should_be_ok():
    pecker = woodpecker(path=".//n/text()", reduce=reducers.first)
    data = pecker(people_root)
    assert data == 'John Smith'


def test_peck_known_predefined_reducer_by_name_should_be_ok():
    pecker = woodpecker(path=".//n/text()", reducer="first")
    data = pecker(people_root)
    assert data == 'John Smith'


def test_peck_callable_reducer_should_be_ok():
    pecker = woodpecker(path=".//n/text()", reduce=lambda xs: xs[1])
    data = pecker(people_root)
    assert data == 'Jane Doe'


def test_peck_callable_reducer_should_take_precedence():
    pecker = woodpecker(path=".//n/text()", reducer="first", reduce=lambda xs: xs[1])
    data = pecker(people_root)
    assert data == 'Jane Doe'


def test_peck_non_matching_path_should_return_none():
    pecker = woodpecker(path=".//foo/text()", reducer="first")
    data = pecker(people_root)
    assert data is None


########################################
# extraction tests
########################################


shining_file = os.path.join(os.path.dirname(__file__), '..', 'examples', 'shining.html')
shining = build_tree(open(shining_file).read())


def test_extract_no_rules_should_return_empty_result():
    data = extract(shining, [])
    assert data == {}


def test_extract_unknown_value_generator_should_raise_error():
    items = [{"key": "title", "value": {"foo": ".//title/text()"}}]
    with raises(ValueError):
        extract(shining, items)


def test_extract_path_value_generator_with_missing_reducer_should_raise_error():
    items = [{"key": "title", "value": {"path": ".//title/text()"}}]
    with raises(ValueError):
        extract(shining, items)


def test_extract_reducer_by_lambda_should_be_ok():
    items = [{"key": "title", "value": {"path": ".//title/text()", "reduce": lambda xs: xs[0]}}]
    data = extract(shining, items)
    assert data == {'title': 'The Shining'}


def test_extract_predefined_reducer_should_be_ok():
    items = [{"key": "title", "value": {"path": ".//title/text()", "reduce": reducers.first}}]
    data = extract(shining, items)
    assert data == {'title': 'The Shining'}


def test_extract_predefined_reducer_by_name_first_should_be_ok():
    items = [{"key": "title", "value": {"path": ".//title/text()", "reducer": "first"}}]
    data = extract(shining, items)
    assert data == {'title': 'The Shining'}


def test_extract_predefined_reducer_by_name_join_should_be_ok():
    items = [{"key": "full_title", "value": {"path": ".//h1//text()", "reducer": "join"}}]
    data = extract(shining, items)
    assert data == {'full_title': 'The Shining (1980)'}


def test_extract_predefined_reducer_by_name_clean_should_be_ok():
    items = [{"key": "review",
              "value": {"path": ".//div[@class='review']//text()", "reducer": "clean"}}]
    data = extract(shining, items)
    assert data == {'review': 'Fantastic movie. Definitely recommended.'}


def test_extract_multiple_rules_should_generate_multiple_items():
    items = [{"key": "title",
              "value": {"path": ".//title/text()", "reduce": reducers.first}},
             {"key": "year",
              "value": {"path": ".//span[@class='year']/text()", "reduce": reducers.first}}]
    data = extract(shining, items)
    assert data == {'title': 'The Shining', 'year': '1980'}


def test_extract_items_with_no_data_should_be_excluded():
    items = [{"key": "title",
              "value": {"path": ".//title/text()", "reduce": reducers.first}},
             {"key": "foo",
              "value": {"path": ".//foo/text()", "reduce": reducers.first}}]
    data = extract(shining, items)
    assert data == {'title': 'The Shining'}


def test_extract_transforming_should_be_applied_after_reducing():
    items = [{"key": "year",
              "value": {"path": ".//span[@class='year']/text()",
                        "reduce": reducers.first,
                        "transform": int}}]
    data = extract(shining, items)
    assert data == {'year': 1980}


def test_extract_multivalued_item_should_be_list():
    items = [{"key": "genres",
              "value": {"foreach": ".//ul[@class='genres']/li",
                        "path": "./text()",
                        "reduce": reducers.first}}]
    data = extract(shining, items)
    assert data == {'genres': ['Horror', 'Drama']}


def test_extract_multivalued_item_empty_should_be_excluded():
    items = [{"key": "foos",
              "value": {"foreach": ".//ul[@class='foos']/li",
                        "path": "./text()",
                        "reduce": reducers.first}}]
    data = extract(shining, items)
    assert data == {}


def test_extract_subrules_should_generate_submappings():
    items = [{"key": "director",
              "value": {
                  "items": [{"key": "name",
                             "value": {"path": ".//div[@class='director']//a/text()",
                                       "reduce": reducers.first}},
                            {"key": "link",
                             "value": {"path": ".//div[@class='director']//a/@href",
                                       "reduce": reducers.first}}]
              }}]
    data = extract(shining, items)
    assert data == {'director': {'link': '/people/1', 'name': 'Stanley Kubrick'}}


def test_extract_multivalued_subrules_should_be_allowed():
    items = [{"key": "cast",
              "value": {
                  "foreach": ".//table[@class='cast']/tr",
                  "items": [
                      {"key": "name",
                       "value": {"path": "./td[1]/a/text()",
                                 "reduce": reducers.first}},
                      {"key": "link",
                       "value": {"path": "./td[1]/a/@href",
                                 "reduce": reducers.first}},
                      {"key": "character",
                       "value": {"path": "./td[2]/text()",
                                 "reduce": reducers.first}}
                  ]
              }}]
    data = extract(shining, items)
    assert data == {'cast': [
        {'character': 'Jack Torrance', 'link': '/people/2', 'name': 'Jack Nicholson'},
        {'character': 'Wendy Torrance', 'link': '/people/3', 'name': 'Shelley Duvall'}
    ]}


def test_extract_subitems_should_be_transformable():
    items = [{"key": "cast",
              "value": {
                  "foreach": ".//table[@class='cast']/tr",
                  "items": [
                      {"key": "name",
                       "value": {"path": "./td[1]/a/text()",
                                 "reduce": reducers.first}},
                      {"key": "character",
                       "value": {"path": "./td[2]/text()",
                                 "reduce": reducers.first}}
                  ],
                  "transform": lambda x: x.get('name') + ' as ' + x.get('character')
              }}]
    data = extract(shining, items)
    assert data == {'cast': ['Jack Nicholson as Jack Torrance',
                             'Shelley Duvall as Wendy Torrance']}


def test_extract_generated_key_path_with_missing_reducer_should_raise_error():
    items = [
        {
            "foreach": ".//div[@class='info']",
            "key": {"path": "./h3/text()"},
            "value": {"path": "./p/text()",
                      "reduce": reducers.first}
        }
    ]
    with raises(ValueError):
        extract(shining, items)


def test_extract_generated_key_path_and_reducer_should_be_ok():
    items = [
        {
            "foreach": ".//div[@class='info']",
            "key": {"path": "./h3/text()",
                    "reduce": reducers.first},
            "value": {"path": "./p/text()",
                      "reduce": reducers.first}
        }
    ]
    data = extract(shining, items)
    assert data == {'Language:': 'English', 'Runtime:': '144 minutes'}


def test_extract_generated_key_normalize_reducer_should_be_ok():
    items = [
        {
            "foreach": ".//div[@class='info']",
            "key": {"path": "./h3/text()",
                    "reduce": reducers.normalize},
            "value": {"path": "./p/text()",
                      "reduce": reducers.first}
        }
    ]
    data = extract(shining, items)
    assert data == {'language': 'English', 'runtime': '144 minutes'}


def test_extract_generated_key_should_be_transformable():
    items = [
        {
            "foreach": ".//div[@class='info']",
            "key": {"path": "./h3/text()",
                    "reduce": reducers.normalize,
                    "transform": lambda x: x.upper()},
            "value": {"path": "./p/text()",
                      "reduce": reducers.first}
        }
    ]
    data = extract(shining, items)
    assert data == {'LANGUAGE': 'English', 'RUNTIME': '144 minutes'}



from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import mark, raises

from io import StringIO

import json
import os

import piculet

if not piculet.PY3:
    from mock import patch
else:
    from unittest.mock import patch


base_dir = os.path.dirname(__file__)
infile = os.path.join(base_dir, 'files', 'utf-8_charset_utf-8.html')
wikipedia_spec = os.path.join(base_dir, '..', 'examples', 'wikipedia.json')


def test_help_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        piculet.main(argv=['piculet', '--help'])
    out, err = capsys.readouterr()
    assert out.startswith('usage: ')


def test_no_command_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        piculet.main(argv=['piculet'])
    out, err = capsys.readouterr()
    assert err.startswith('usage: ')
    assert ('required: command' in err) or ('too few arguments' in err)


def test_invalid_command_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        piculet.main(argv=['piculet', 'foo'])
    out, err = capsys.readouterr()
    assert err.startswith('usage: ')
    assert ('invalid choice: \'foo\'' in err) or ('invalid choice: u\'foo\'' in err)


def test_unrecognized_arguments_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        piculet.main(argv=['piculet', '--foo', 'h2x', ''])
    out, err = capsys.readouterr()
    assert err.startswith('usage: ')
    assert 'unrecognized arguments: --foo' in err


def test_debug_mode_should_print_debug_messages_on_stderr(capsys):
    piculet.main(argv=['piculet', '--debug', 'h2x', infile])
    out, err = capsys.readouterr()
    assert 'running in debug mode' in err


def test_h2x_no_input_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        piculet.main(argv=['piculet', 'h2x'])
    out, err = capsys.readouterr()
    assert err.startswith('usage: ')
    assert ('required: file' in err) or ('too few arguments' in err)


def test_h2x_should_read_given_file(capsys):
    with open(infile, 'rb') as f:
        content = piculet.decode_html(f.read())
    piculet.main(argv=['piculet', 'h2x', infile])
    out, err = capsys.readouterr()
    assert out == content


def test_h2x_should_read_stdin_when_input_is_dash(capsys):
    with open(infile, 'rb') as f:
        content = piculet.decode_html(f.read())
    with patch('sys.stdin', StringIO(content)):
        piculet.main(argv=['piculet', 'h2x', '-'])
    out, err = capsys.readouterr()
    assert out == content


def test_scrape_no_url_should_print_usage_and_exit(capsys):
    with raises(SystemExit):
        piculet.main(argv=['piculet', 'scrape', '-s', wikipedia_spec])
    out, err = capsys.readouterr()
    assert err.startswith('usage: ')
    assert ('required: document' in err) or ('too few arguments' in err)


def test_scrape_no_spec_should_print_usage_and_exit(capsys, test_pages):
    url, _ = test_pages['bowie']
    with raises(SystemExit):
        piculet.main(argv=['piculet', 'scrape', url])
    out, err = capsys.readouterr()
    assert err.startswith('usage: ')
    assert ('required: -s' in err) or ('--spec is required' in err)


def test_scrape_missing_spec_file_should_fail_and_exit(capsys, test_pages):
    url, _ = test_pages['bowie']
    with raises(SystemExit):
        piculet.main(argv=['piculet', 'scrape', url, '-s', 'foo.json'])
    out, err = capsys.readouterr()
    assert 'No such file or directory: ' in err


def test_scrape_local_should_scrape_given_file(capsys):
    dirname = os.path.join(os.path.dirname(__file__), '..', 'examples')
    shining = os.path.join(dirname, 'shining.html')
    spec = os.path.join(dirname, 'movie.json')
    piculet.main(argv=['piculet', 'scrape', shining, '-s', spec])
    out, err = capsys.readouterr()
    data = json.loads(out)
    assert data['title'] == 'The Shining'


def test_scrape_should_scrape_given_url(capsys, test_pages):
    url, _ = test_pages['bowie']
    piculet.main(argv=['piculet', 'scrape', url, '-s', wikipedia_spec, '--html'])
    out, err = capsys.readouterr()
    data = json.loads(out)
    assert data['name'] == 'David Bowie'


def test_scrape_cached_should_read_from_disk(capsys, test_pages):
    url, cache = test_pages['bowie']
    assert os.path.exists(cache)
    piculet.main(argv=['piculet', 'scrape', url, '-s', wikipedia_spec, '--html'])
    out, err = capsys.readouterr()
    data = json.loads(out)
    assert data['name'] == 'David Bowie'


@mark.download
def test_scrape_cache_disabled_should_retrieve_from_web(capsys, test_pages, enable_internet):
    cache_dir = os.environ['PICULET_WEB_CACHE']  # backup cache dir
    del os.environ['PICULET_WEB_CACHE']
    url, cache = test_pages['bowie']
    assert os.path.exists(cache)
    os.rename(cache, cache + '.BACKUP')
    try:
        piculet.main(argv=['piculet', 'scrape', url, '-s', wikipedia_spec, '--html'])
        assert not os.path.exists(cache)
        out, err = capsys.readouterr()
        data = json.loads(out)
        assert data['name'] == 'David Bowie'
    finally:
        os.rename(cache + '.BACKUP', cache)
        os.environ['PICULET_WEB_CACHE'] = cache_dir  # restore cache dir


@mark.download
def test_scrape_uncached_should_retrieve_from_web(capsys, test_pages, enable_internet):
    url, cache = test_pages['bowie']
    assert os.path.exists(cache)
    os.rename(cache, cache + '.BACKUP')
    try:
        piculet.main(argv=['piculet', 'scrape', url, '-s', wikipedia_spec, '--html'])
        assert os.path.exists(cache)
        out, err = capsys.readouterr()
        data = json.loads(out)
        assert data['name'] == 'David Bowie'
    finally:
        os.rename(cache + '.BACKUP', cache)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from pytest import raises

import os

from piculet import decode_html, html_to_xhtml


def read_document(filename):
    path = os.path.join(os.path.dirname(__file__), 'files', filename)
    with open(path, 'rb') as f:
        content = f.read()
    return content


def test_decode_content_meta_charset_correct_should_succeed():
    content = read_document('utf-8_charset_utf-8.html')
    assert 'ğışĞİŞ' in decode_html(content)


def test_decode_content_meta_content_type_correct_should_succeed():
    content = read_document('utf-8_content-type_utf-8.html')
    assert 'ğışĞİŞ' in decode_html(content)


def test_decode_content_meta_charset_incorrect_should_fail():
    content = read_document('utf-8_charset_iso8859-9.html')
    assert 'ğışĞİŞ' not in decode_html(content)


def test_decode_content_meta_content_type_incorrect_should_fail():
    content = read_document('utf-8_content-type_iso8859-9.html')
    assert 'ğışĞİŞ' not in decode_html(content)


def test_decode_content_meta_charset_incompatible_should_raise_unicode_error():
    content = read_document('iso8859-9_charset_utf-8.html')
    with raises(UnicodeDecodeError):
        decode_html(content)


def test_decode_content_meta_content_type_incompatible_should_raise_unicode_error():
    content = read_document('iso8859-9_content-type_utf-8.html')
    with raises(UnicodeDecodeError):
        decode_html(content)


def test_decode_content_requested_charset_correct_should_succeed():
    content = read_document('utf-8_charset_iso8859-9.html')
    assert 'ğışĞİŞ' in decode_html(content, charset='utf-8')


def test_decode_content_requested_charset_incorrect_should_fail():
    content = read_document('utf-8_charset_utf-8.html')
    assert 'ğışĞİŞ' not in decode_html(content, charset='iso8859-9')


def test_decode_content_requested_charset_incompatible_should_raise_unicode_error():
    content = read_document('iso8859-9_charset_iso8859-9.html')
    with raises(UnicodeDecodeError):
        decode_html(content, charset='utf-8')


def test_decode_content_fallback_default_correct_should_succeed():
    content = read_document('utf-8_charset_none.html')
    assert 'ğışĞİŞ' in decode_html(content)


def test_decode_content_fallback_correct_should_succeed():
    content = read_document('iso8859-9_charset_none.html')
    assert 'ğışĞİŞ' in decode_html(content, fallback_charset='iso8859-9')


def test_decode_content_fallback_incorrect_should_fail():
    content = read_document('iso8859-9_charset_none.html')
    assert 'ğışĞİŞ' not in decode_html(content, fallback_charset='iso8859-1')


def test_decode_content_fallback_incompatible_should_raise_unicode_error():
    content = read_document('iso8859-9_charset_none.html')
    with raises(UnicodeDecodeError):
        decode_html(content, fallback_charset='utf-8')


def test_html_to_xhtml_well_formed_xml_should_succeed():
    content = '''<html></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html></html>'''


def test_html_to_xhtml_doctype_should_be_removed():
    content = '''<!DOCTYPE html><html></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html></html>'''


def test_html_to_xhtml_comment_should_be_removed():
    content = '''<html><!-- comment --></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html></html>'''


def test_html_to_xhtml_omitted_tags_should_be_removed():
    content = '''<html><p></p><font></font></html>'''
    normalized = html_to_xhtml(content, omit_tags={'font'})
    assert normalized == '''<html><p></p></html>'''


def test_html_to_xhtml_omitted_attributes_should_be_removed():
    content = '''<html><p font="times"></p></html>'''
    normalized = html_to_xhtml(content, omit_attrs={'font'})
    assert normalized == '''<html><p></p></html>'''


def test_html_to_xhtml_self_closing_tags_should_have_slash_at_end():
    content = '''<html><br></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><br /></html>'''


def test_html_to_xhtml_attributes_should_have_values():
    content = '''<html><option checked></option></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><option checked=""></option></html>'''


def test_html_to_xhtml_unclosed_tags_should_be_closed():
    content = '''<html><p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p></p></html>'''


def test_html_to_xhtml_unclosed_lis_should_be_closed():
    content = '''<html><ul><li><li></li></ul></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><ul><li></li><li></li></ul></html>'''


def test_html_to_xhtml_unclosed_last_lis_should_be_closed():
    content = '''<html><ul><li></li><li></ul></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><ul><li></li><li></li></ul></html>'''


def test_html_to_xhtml_end_tag_without_start_tag_should_be_discarded():
    content = '''<html><p></p></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p></p></html>'''


def test_html_to_xhtml_incorrect_nesting_should_be_reordered():
    content = '''<html><div><p></div></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><div><p></p></div></html>'''


def test_html_to_xhtml_ampersands_should_be_replaced_in_data():
    content = '''<html><p>&</p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p>&amp;</p></html>'''


def test_html_to_xhtml_lts_should_be_replaced_in_data():
    content = '''<html><p><</p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p>&lt;</p></html>'''


def test_html_to_xhtml_gts_should_be_replaced_in_data():
    content = '''<html><p>></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p>&gt;</p></html>'''


def test_html_to_xhtml_ampersands_should_be_replaced_in_attribute_values():
    content = '''<html><p id="&"></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p id="&amp;"></p></html>'''


def test_html_to_xhtml_lts_should_be_replaced_in_attribute_values():
    content = '''<html><p id="<"></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p id="&lt;"></p></html>'''


def test_html_to_xhtml_gts_should_be_replaced_in_attribute_values():
    content = '''<html><p id=">"></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p id="&gt;"></p></html>'''


def test_html_to_xhtml_quotes_should_be_replaced_in_attribute_values():
    content = '''<html><p id="&#x22;"></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p id="&quot;"></p></html>'''


def test_html_to_xhtml_unicode_data_should_be_preserved():
    content = '''<html><p>ğış</p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p>ğış</p></html>'''


def test_html_to_xhtml_unicode_attribute_value_should_be_preserved():
    content = '''<html><p foo="ğış"></p></html>'''
    normalized = html_to_xhtml(content)
    assert normalized == '''<html><p foo="ğış"></p></html>'''



from setuptools import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

setup(
    name='piculet',
    version='1.0b2',
    description='XML/HTML scraper using XPath queries.',
    long_description=readme + '\n\n' + history,
    url='https://bitbucket.org/uyar/piculet',
    author='H. Turgut Uyar',
    author_email='uyar@tekir.org',
    license='LGPL',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Utilities'
    ],
    keywords='xml html xpath scrape json',
    py_modules=['piculet'],
    install_requires=[],
    extras_require={
        'dev': [
            'flake8',
            'wheel',
            'twine'
        ],
        'doc': [
            'sphinx',
            'sphinx_rtd_theme',
            'pygenstub'
        ],
        'test': [
            'pytest',
            'pytest-cov'
        ],
    },
    entry_points="""
        [console_scripts]
        piculet=piculet:main
    """
)
